
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
/*
 * #/include <sys/malloc.h>
 */
#include <limits.h>
#include <time.h>
#include <sys/types.h>
#include <sys/times.h>
#include <string.h>
#include <stdlib.h>

/*
 *   Where the error messages go to (usually stderr, or stdout for web use)
 */

#define WARNING_STREAM stdout

/*
* Constant definitions
*/

#define F_MAXINT 1000000000
#define REF_MAX 10000
#define REF_DEFAULT 10000
#define EXTRA_DEFAULT 500
#define MAX_REF_SIZE 8
#define STAK_MAX (REF_MAX*MAX_REF_SIZE)
#define MAX_PENALTY F_MAXINT

/* This is mandatory (set it to the right value) if CLOCKS_PER_SEC is
   set to the dafault value of a million, since that probably means
   the local value of the clock tick could not be found.
*/
#define LOCAL_CLK_TCK 100

/*
 * WSZ (word size) is the number of bits in a "logical" variable.
 */

#define WSZ 32

#define SURJ_MAX 256
#define RRA 50
#define RRB 30

#define NO_REPORT ((unsigned)1)
#define INT_REPORT ((unsigned)2)
#define FLOAT_REPORT ((unsigned)4)
#define DO_REPORT ((unsigned)8)

#define logical unsigned int

#define FOR(x,y) for( y = x; y; y = y->p )
#define FORR(x,y) FOR( x->cellvals, y )

/*
* Type definitions
*/

#define boolean char
#define false ((boolean)0)
#define true ((boolean)1)

typedef logical *set;

typedef enum {
  NOCUT,
  FOUND,
  RECORDED,
  COMPLETED
} cutcode;

typedef enum {
  PERFECT,
  NULL_SPACE,
  BIG_SPACE,
  REF_LIST,
  BUST,
  IBUST,
  NULL_REF,
  SURJES,
  SKIP
} exit_code;

typedef enum {
  SETPOSSES,
  DOSMALLS,
  SEARCHES,

  SOLUTIONS,
  BAD_TESTS,
  SURJ_BACKTRACKS,
  NULL_BACKTRACKS,

  VECMAX,
  INVALUES,
  PREREFS,
  REFSIZE,
  TOPLESS,

  LOC_MEMORY,
  MEMORY,

  F_SUBSUMED,
  B_SUBSUMED,
  B1_SUBSUMED,
  SOFT_DOUBLE,

  BRANCHES,

  PRE_TIME,
  SEARCH_TIME,

  STOP_CODE,
  MAX_REPORT
} report_code;

typedef struct IL {
  boolean positive;             /* Represents positive literal */
  short int v, w;               /* A couple of integers */
  struct IL *p;                 /* in a list */
} INTLIST, *intlist;


typedef struct reftype {
  int cardinality;             /* Just what is says */
  boolean notop;               /* Useless extension omitted */
  intlist cellvals;            /* The cell-valuess involved */
  int softweight;              /* Cost if violated if a soft constraint */
  struct reftype *nxtr;        /* Just for stack management */
} REF, *ref;


typedef struct rf_list {        /* A list of refutations  */
  ref rf;
  struct rf_list *nextref;
} REFLIST, *reflist;


typedef struct {
  int blocks;                 /* How many times blocked */
  ref back;                   /* For calculating secondaries */
  reflist cvrefs;             /* The refs involving this c-v */
  reflist posrefs;            /* Positives involving this c-v */
  int vpos;                   /* Offset of its cell */
  int penalty;                /* Soft cost of inserting this value */
} CELLVALUE, *cellvalue;


typedef struct {
  short poss;                 /* How many possible values */
  short csiz;                 /* Cardinality of {0, ... max value} */
  short vpos;                 /* Vector position of this cell */
  reflist inject;             /* Index to injective function */
  CELLVALUE *cvals;           /* The cell-value structures (array) */
  boolean used;               /* Involved in new refutation */
  boolean has_value;          /* This cell has a value */
  boolean fixed;              /* Has only one possible value */
  int times_picked;           /* How often this cell split upon */
  int least_penalty;          /* Penalty of the best available value */
} CELL, *cell;


typedef struct {
  set valset;                 /* Values which must occur somewhere */
  int *last_seen;             /* Which cell witnesses value (array 0..siz) */
  intlist cellset;            /* Where the values must occur */
} SURJ;


typedef struct {
  logical report_type;
  char s_report[80];
  int i_report;
  float f_report;
  void (*do_report)();
} REPORT;


typedef struct cng {
  int forkpoint;             /* Cell where division takes place */
  int cutval;                /* Value for that cell */
  int *icomvec;              /* Current values of all (significant) cells */
  int *inorder;              /* Order in which they should be placed */
  set *possvals;             /* Domains including blocked values */
  set *valsleft;             /* Values waiting to be tried */
  boolean active;            /* Cut is currently being applied */
  struct cng *lastone;
} CUTT, *cutt;

#define NEWSLENGTH 1024

typedef struct { REF news[NEWSLENGTH]; }  RBATCH,  *rbatch;
typedef struct { INTLIST news[NEWSLENGTH]; }  ILBATCH, *ilbatch;
typedef struct { REFLIST news[NEWSLENGTH]; }  RLBATCH, *rlbatch;


typedef struct {
  boolean (*Test)();            /* The principal test function  */
  boolean (*Onerefs)();         /* Removal of impossible values */
  boolean (*Smallrefs)();       /* Special treatment for smalls */
  void (*Sizes)();              /* Get the domain sizes */
  int vlength;                  /* Local vector length */
  int maxval;                   /* Local max domain size */
  int batches;                  /* Local stack size / 1000 */
  int extra_batches;            /* Batches after pretest */
  int maxref;                   /* Maximum refutation size */
  int maxbak;                   /* Maximum number of backtracks */
  int maxinvals;                /* Maximum number of value insertions */
  int verb;                     /* Statistics verbosity setting */
  int ref_record[RRA][RRB];
  int *branches;                /* For stats (aray 0..siz) */
  int max_penalty;              /* Maximum badness allowable */
  int dump_rate;                /* How often to dump a trace */
  int tracepoints;              /* How often trace dump point reached */
  boolean piping;               /* Some functions come from the pipe */
  boolean Nosecs;               /* Suppress secondary refs */
  boolean topsoff;              /* Remove redundant ref tops */
  boolean init_sets_ok;         /* Remove redundant ref tops */
  boolean report_breaks;        /* Print reason for end search */
  boolean print_headers;        /* Print labels "Model N" etc */
  boolean dumpdepth;            /* Print backtracking depths */
  boolean awnty_enabled;        /* Print search tree estimate data */
  boolean cgv_enabled;          /* Print data for search visualisation */
  boolean *forbidden;           /* No backtracking (array 0..vlength) */
  boolean random;               /* Insert values in random order */
  boolean done;                 /* End-of-search status raised */
  boolean profile;              /* Print info for visualisation  */
  REPORT  reps[MAX_REPORT+1];   /* Search statistics */
} TRIN, *trin;



typedef struct {
  trin job;
  cell vector;                  /* The decision variables (array) */
  set *scomvec;                 /* Set communication vector (array) */
                                /* - during search, remaining poss values */
  int *icomvec;                 /* Integer communication vector (array) */
                                /* - during search, curent values */
  int *inorder;                 /* When values were inserted (array) */
  int *coinorder;               /* Converse of inorder (array) */
  int *priority;                /* Level of function declarations (array) */
  int *best;                    /* Lowest penalty possible for cell (array) */
  int best_penalty;             /* Lower bound on sum of penalties */
  int total_penalty;            /* Actual penalty for inserted (array) values */
  int *awnty_past;              /* Number of subtrees searched (array) */
  int *awnty_future;            /* Number (possibly) to be searched (array) */
  int *awnty_sofar;             /* Cumulative subtree size */
  int this_cell;                /* Ordinal of current insertion */
  int susguy;                   /* Wiped out cell (for finding nogood) */
  int firstchange;              /* Most significant change since last model */
  int forkpoint;                /* Most significant really branching node */
  int maxval;                   /* Max domain size for any cell */
  int stakmax;                  /* Max batches * max ref size */
  int unfixt;                   /* Highest numbered cell with no value */
  int local_memory;             /* Current size of constraint database */
  int max_batches;              /* Bounds number of batches of constraints */
  boolean novect;               /* "No vector": no solution exists */
  boolean ts;                   /* "Totally suspended": wipeout detecte */
  boolean cut_recorded;         /* Search space division has been processed */
  boolean noref;                /* Nothing (?) - appears to be dead code */
  rbatch stak[REF_MAX];                  /* Reservoir topup chunks */
  ilbatch il_stak[STAK_MAX];             /* Reservoir topup chunks */
  rlbatch rl_stak[STAK_MAX + REF_MAX];   /* Reservoir topup chunks */
  ref stak_ptr;                          /* Reservoir */
  intlist il_free;                       /* Reservoir */
  reflist rl_free;                       /* Reservoir */
  reflist duds;                 /* Expired constraints (marked for deletion) */
  ref newref;                   /* Next constraint to be incorporated */
  SURJ surjection[SURJ_MAX];    /* Sets of cells in surjective functions */
  cutt cutptr;                  /* List of search space divisions */
  cutcode cut_flag;             /* Status of search space division */
  boolean absolute;             /* No real branching has happened yet */
} TRS, *trs;



/*
*  Global variables
*/

#ifdef TRFILE
#define TR_GLOB
#else
#define TR_GLOB extern
#endif

//---------------------- Annoying but needed ---------------------------//

TR_GLOB int theClockTick;

TR_GLOB trs debugger;
TR_GLOB int begin_time, end_time;

// global variable to handle exit on time_limit
TR_GLOB volatile boolean bug_out;

TR_GLOB int *debug_int;

/*
*  Function prototypes
*/


void transref(trin job);
void free_trs(trs T);
void init_trin(trin tr);
void clean_up(trs T);
void report_initial(trs T);
trs tr_initial(trin job);
int search(trs T, int dpth, int bdpth);
int remaining(cell c, int wyord[], int bound);
int pick_cell(trs T);
int set_fork( trs T, int rc);
void randomise(int arr[], int bound);
void greedy_order(int arr[], int bound, int pos, cell c, trs T);
void delete_cut(trs T);
reflist get_rl(trs T);
void put_rl(trs T, reflist dead);
intlist get_il(trs T, boolean b, int x, int y, intlist ip);
void put_il(trs T, intlist dead);
int batches_used(trs T);
ref get_ref(trs T);
void put_ref(trs T, ref dead, boolean flag);
void rec_put_il(trs T, intlist ip, ref r, boolean flag);
void more_refs(trs T);
void more_rls(trs T);
void more_ils(trs T);
boolean invalue(trs T, cell c, int v);
void transfer_ref (trs T, ref rf);
void blockval(trs T, cell c, int v, ref rrf, report_code RC);
boolean surjok(trs T);
void find_min_penalty(cell c, trs T);
void set_total_penalty(trs T);
void outvalue(trs T, cell c);
void untransfer_ref (trs T, cell c, ref rf);
void unblockval(trs T, cell c, int v, ref rrf);
void mark_delete(trs T, ref r);
void delete_duds(trs T);
void delete_a_dud(trs T, reflist r);
void setref(trs T);
void reduceref(trs T);
void addref(int x, trs T);
void totsus(trs T);
void no_topless_singles(trs T);
void make_coherent(trs T);
void no_ref(trs T);
void new_small_ref(int cells[], int vals[], boolean valency[],
		   int softweight, trs T);
void remove_value(trs T, cell c, int x);
intlist get_all_ils(int cells[], int vals[], boolean valency[], int x, trs T);
boolean subsumed(ref the_ref, trs T);
boolean subsumes(ref little_ref, ref big_ref);
void back_sub(trs T, ref the_ref);
boolean identical_ref( ref the_ref, trs T );
boolean is_the_same( ref the_new_ref, ref the_old_ref );
void Ref(int x, trs T);
void record_cut(trs T);
void record_part_cut(set s, trs T);
void complete_cut(trs T);
boolean prepared(trs T);
void setinx(trs T);
boolean pre_test(trs T);
boolean absfixt(int x);
void surjective(boolean vec[], set vset, trs T);
void injective(boolean vec[], int x, trs T);
void co_priority(int x, int y, trs T);
void ref_size_report();
void report_branches(trin tr);
void stopped_by( trin tr );
void dump_refs(int rows, int columns);
void job_done(trin t, char *s);
int timestamp();
int timestat(trin Job, int old_t, report_code rc);
void skipout(char s[], int exitcode);
void add_penalty(int vp, int val, int x, trs T);
void subtract_penalty(int vp, int val, int x, trs T);
void set_best(trs T);
void dump_depth(int d);
void dump_awnty(int past[], int future[], int sofar[], int depth, boolean term);

/*
 * Prototypes of functions in sets.c
 */

boolean init_sets_needed(int x);
void init_sets(int x);
set new_set();
set fullset(set s);
set emptyset(set s);
boolean isemptyset(set s);
set singleton(int x, set s);
boolean issingle(set s);
set assignset(set dest, set source);
void intoset(int x, set s);
void outofset(int x, set s);
boolean isinset(int x, set s);
void intersect(set dest, set mask);
void freeset(set s);
void really_free_sets(void);

/*
 * These in fprepare.c
 */

void print_un_vec_pos(int x);
void print_un_vec_value(int x, int val);
void report_constraint(ref thisref);

/*
 * These in foutput.c
 */

void dump_trace(trin tr, boolean final);
void print_constraint_graph(trs T);
void check_new_ref(int x, ref rf, trs T);
void cgv_state_report(trs T, cell c, int val);
void dump_state(set domains[], trs T);
void print_cgv_set(set domain, int maxval);
