#include "diagnose.h"

static int theSeed;
static int theTimelimit;

dclause debugger_dc;

boolean too_long_to_be_soft();


int main(int argc, char *argv[])
{
  clauselist clauses;            /* Clauses (my data structure) */
  dialogue history = 0;          /* The story so far */
  char tempfilename[16];         /* infil parsed and annotated */
  char sorts_and_functions[16];  /* Problem without the clauses */
  char command[SLEN];

  if (argc!=2) {
    fprintf(stderr,"  synopsis: diagnose <file name>\n");
    exit(2);
  }
  theSeed = defaultSeed;
  theTimelimit = defaultTimelimit;
  nonce_filename(tempfilename);
  nonce_filename(sorts_and_functions);
  init_init_finder();
  set_job_only();
  transfer_job(argv[1],tempfilename);
  unset_job_only();
  get_problem_basis(tempfilename,sorts_and_functions);
  clauses = get_problem_clauses(tempfilename);
  sprintf(command,"rm -f %s",tempfilename);
  system(command);
  do {
    generate_diagnosis(sorts_and_functions,clauses);
    history = next_action(clauses,history);
  }
  while (!history || history->e != halt);
  sprintf(command,"rm -f %s",sorts_and_functions);
  system(command);
  return 0;
}


void nonce_filename(char path[])
{
  int x;
  int r;

  sprintf(path+10,".temp");
  do {
    for (x=0; x<10; x++) {
      r = lrand48() % 62;
      if (r < 10) path[x] = '0'+r;
      else if (r < 36) path[x] = 'A'+(r-10);
      else path[x] = 'a'+(r-36);
    }
  }
  while (file_exists(path));
}

boolean file_exists(char *path)
{
  FILE *f;

  f = fopen(path,"r");
  if (f) {
    fclose(f);
    return true;
  }
  return false;
}


void get_problem_basis(char *source, char *dest)
{
  FILE *fin;
  FILE *fout;
  char line[1000];
  char keyword[100];
  int x;

  fin = fopen(source,"r");
  if (!fin) inerror("Failed to open input file",source);
  fout = fopen(dest,"w");
  if (!fout) inerror("Failed to open output file",dest);
  while (fgets(line,1000,fin)) {
    for (x=0; line[x] && isspace(line[x]); x++) ;
    if (line[x]) {
      sscanf(line,"%s",keyword);
      if (!strcmp(keyword,"sort") ||
	  !strcmp(keyword,"function")) {
	fputs(line,fout);
	while (!strchr(line,'}')) {
	  fgets(line,1000,fin);
	  fputs(line,fout);
	}
      }
    }
  }
  fclose(fin);
  fclose(fout);
}


clauselist get_problem_clauses(char *filename)
{
  clauselist cl = 0;
  clauselist new;
  char line[1000];
  char keyword[100];
  FILE *fin;
  int x;

  fin = fopen(filename,"r");
  if (!fin) inerror("Failed to open input file",filename);
  while (fgets(line,1000,fin)) {
    for (x=0; line[x] && isspace(line[x]); x++) ;
    if (line[x]) {
      sscanf(line,"%s",keyword);
      if (!strcmp(keyword,"clause")) {
	new = (clauselist) malloc(sizeof(CLAUSELIST));
	new->next = cl;
	new->last = 0;
	if (cl) cl->last = new;
	cl = new;
	cl->c = get_dclause(line);
      }
    }
  }
  fclose(fin);
  return cl;
}


dclause get_dclause(char *line)
{
  int x;
  int y = 0;
  int z;
  dclause rc;
  char temp[1000];

  for (x=0; x<1000; x++) temp[x] = 0;
  for (x=0; line[x] && line[x] != '{'; x++) ;
  for (x++; line[x] && line[x] != '}'; x++) temp[y++] = line[x];
  while (line[x] != '%') x++; x++;
  rc = (dclause) malloc(sizeof(DCLAUSE));
  dynstring_initialise(&(rc->s),temp);
  sscanf(line+x,"%d%d%d%d",
	 &(rc->var_count),
	 &(rc->weight),
	 &(rc->softweight),
	 &(rc->id));
  for (y=0; y<4; y++) {
    while (line[x] && isspace(line[x])) x++;
    while (line[x] && !isspace(line[x])) x++;
  }
  if (rc->var_count) {
    rc->var_name = (char**) malloc(sizeof(char)*rc->var_count);
    rc->var_sort = (char**) malloc(sizeof(char)*rc->var_count);
    for (y=0; y<rc->var_count; y++) {
      while (isspace(line[x])) x++;
      for (z=0; !isspace(line[x+z]); z++);
      rc->var_name[y] = (char*) malloc((sizeof(char)*z)+1);
      sscanf(line+x,"%s",rc->var_name[y]);
      while (!isspace(line[x])) x++;
      while (isspace(line[x])) x++;
      for (z=0; !isspace(line[x+z]); z++);
      rc->var_sort[y] = (char*) malloc((sizeof(char)*z)+1);
      sscanf(line+x,"%s",rc->var_sort[y]);
      while (!isspace(line[x])) x++;
    }
  }
  rc->sacred = false;
  rc->satisfied = false;
  return rc;
}



void generate_diagnosis(char *sorts_and_functions, clauselist clauses)
{
  char tempfilename[16];
  FILE *basis;
  FILE *snf;
  char line[1000];
  char command[1000];
  clauselist cptr;
  int good;
  dynstring clause_input;

  snf = fopen(sorts_and_functions,"r");
  if (!snf) inerror("Failed to open input file",sorts_and_functions);
  nonce_filename(tempfilename);
  basis = fopen(tempfilename,"w");
  if (!basis) inerror("Failed to open temporary file",tempfilename);
  while (fgets(line,1000,snf)) fputs(line,basis);
  fclose(snf);
  for (cptr=clauses; cptr; cptr=cptr->next) {
    if (too_long_to_be_soft(cptr->c->weight,cptr->c->var_count))
      cptr->c->sacred = true;
    if (cptr->c->sacred)
      fprintf(basis,"clause { %s }\n", cptr->c->s.c_str);
    else
      fprintf(basis,"clause (1) { %s }\n", cptr->c->s.c_str);
  }
  fprintf(basis,"setting verbosity job none\n");
  fprintf(basis,"setting verbosity models none\n");
  fprintf(basis,"setting verbosity stats full\n");
  fprintf(basis,"setting time_limit %d\n",theTimelimit);
  fprintf(basis,"setting random %d\n",theSeed);
  fprintf(basis,"setting no_print_headers\n");
  fprintf(basis,"end\n");
  fclose(basis);
  good = init_finder(tempfilename);
  if (!good)
    report_no_model(clauses);
  else {
    for (cptr=clauses; cptr; cptr=cptr->next)
      if (cptr->c->sacred)
	cptr->c->satisfied = true;
      else {
	dynstring_initialise(&clause_input,"clause { ");
	dynstring_strcat(&clause_input,&(cptr->c->s));
	dynstring_strcatc(&clause_input," }");
	cptr->c->satisfied = found(&clause_input,0,0);
      }
    report_model(clauses);
    theTimelimit = defaultTimelimit;
  }
  sprintf(command,"rm -f %s",tempfilename);
  system(command);
  FINDER_cleanup();
}


void report_no_model(clauselist clauses)
{
  clauselist cptr;
  int sacred_cows = 0;

  for (cptr=clauses; cptr; cptr=cptr->next)
    cptr->c->satisfied = false;
  system("clear");
  for (cptr=clauses; cptr; cptr=cptr->next)
    if (cptr->c->sacred)
      sacred_cows++;
  if (!sacred_cows) {
    printf("The specification of sorts and functions seems inconsistent.\n");
    printf("This needs to be fixed before we can examine the clauses.\n");
  }
  else {
    printf("No model at all was found satisfying ");
    if (sacred_cows == 1) printf("the clause ");
    else if (sacred_cows == 2) printf("both of the clauses ");
    else printf("all %d of the clauses ", sacred_cows);
    printf("\nfixed as \"definitely true\".\n");
  }
}


void report_model(clauselist clauses)
{
  clauselist cptr;
  int x = 0;

  system("clear");
  for (cptr=clauses; cptr; cptr=cptr->next)
    if (!cptr->c->satisfied) {
      if (!x)
	printf("Suggested diagnosis - the following could be false:\n\n");
      printf("%4d%s\n", ++x, cptr->c->s.c_str);
    }
  printf("\nHere is a model that satisfies everything");
  if (x)
    printf(" else");
  printf(":\n");
  print_model_anyway();
}


dialogue next_action(clauselist clauses, dialogue history)
{
  char c;
  char list_of_ints[SLEN];
  clauselist cptr;
  int n = 0;
  int nunsat = 0;
  int nsacred = 0;
  int x, y;
  boolean repeat;
  boolean long_list = false;

  for (cptr=clauses; cptr; cptr=cptr->next) {
    n++;
    if (!cptr->c->satisfied)
      nunsat++;
    if (cptr->c->sacred)
      nsacred++;
  }

  do {
    repeat = false;
    printf("\nWhat do you want to do?\n\n");
    if (nunsat)
      printf("  (a) Fix clause(s) as true\n");
    else
      printf("  (a) [Fix clause(s): unavailable as all clauses are true]\n");
    if (nsacred)
      printf("  (b) Unfix a clause\n");
    else
      printf("  (b) [Unfix a clause: unavailable as no clauses are fixed]\n");
    if (history)
      printf("  (c) Go back to the previous state\n");
    else
      printf("  (c) [Go back: unavailable as there is no previous state]\n");
    printf("  (d) Display current state\n");
    printf("  (e) Search again with longer time limit\n");
    printf("  (x) Exit\n\nAnswer:  ");
    fflush(stdout);
    do c = getchar();
    while (!((c=='a' && nunsat) || 
	     (c=='b' && nsacred) || 
	     (c=='c' && history) ||
	     (c=='d' || c=='e' || c=='x')));
    switch(c) {
    case 'a':
      history = new_dialogue(history,protect);
      if (nunsat==1) strcpy(list_of_ints,"1");
      else {
	printf("\nWhich clause(s) do you want to fix?  ");
	fflush(stdout);
	do
	  fgets(list_of_ints,SLEN,stdin);
	while (!(digit_in(list_of_ints)));
      }
      for (cptr=clauses; cptr; cptr=cptr->next)
	cptr->c->flip = false;
      while ((x = next_int(list_of_ints)) > 0) {
	for (y=0, cptr=clauses; cptr; cptr=cptr->next)
	  if (!((cptr->c->satisfied && !long_list) || cptr->c->sacred)) {
	    y++;
	    if (x == y) {
	      cptr->c->flip = true;
	      break;
	    }
	  }
      }
      for (cptr=clauses; cptr; cptr=cptr->next)
	if (cptr->c->flip) {
	  cptr->c->sacred = true;
	  history->parameters = add_parameter(cptr->c,history->parameters);
	}
      break;
    case 'b':
      history = new_dialogue(history,unprotect);
      if (nsacred==1) {
	printf("Unfixing  ");
	for (cptr=clauses; cptr; cptr=cptr->next)
	  if (cptr->c->sacred) {
	    printf("%s\n", cptr->c->s.c_str);
	    cptr->c->sacred = false;
	    history->parameters = add_parameter(cptr->c,history->parameters);
	    break;
	  }
      }
      else {
	printf("\nCurrently fixed clauses are:\n");
	for (x=0, cptr=clauses; cptr; cptr=cptr->next)
	  if (cptr->c->sacred)
	    printf("%4d%s\n", ++x, cptr->c->s.c_str);
	do {
	  printf("\nWhich clause do you want to unfix?  ");
	  fflush(stdout);
	  scanf("%d",&x);
	}
	while (x < 0 || x > nsacred);
	for (y=0, cptr=clauses; cptr; cptr=cptr->next)
	  if (cptr->c->sacred) {
	    y++;
	    if (x == y) {
	      cptr->c->sacred = false;
	      history->parameters = add_parameter(cptr->c,history->parameters);
	      break;
	    }
	  }
      }
      break;
    case 'c':
      if (history->parent) {
	history = history->parent;
	switch(history->e) {
	case protect:
	  {
	    clauselist cptr;
	    for (cptr=history->parameters; cptr; cptr=cptr->next)
	      cptr->c->sacred = false;
	  }
	  break;
	case unprotect:
	  {
	    clauselist cptr;
	    for (cptr=history->parameters; cptr; cptr=cptr->next)
	      cptr->c->sacred = true;
	  }
	  break;
        case change_time_limit:
	  theTimelimit = history->time;
	  break;
	case halt:
	  printf("What is this? - we are not supposed to be here!\n");
	  break;
	}
	theTimelimit = history->time;
      }
      break;
    case 'd':
      display_current(clauses);
      long_list = true;
      repeat = true;
      break;
    case 'e':
      theTimelimit *= 2;
      theSeed++;
      history = new_dialogue(history,change_time_limit);
      break;
    case 'x':
      history = new_dialogue(history,halt);
      break;
    }
  }
  while (repeat);
  return history;
}


clauselist add_parameter(dclause c, clauselist cl)
{
  clauselist new = (clauselist) malloc(sizeof(CLAUSELIST));
  new->next = cl;
  new->last = 0;
  if (new->next) new->next->last = new;
  new->c = c;
  return new;
}


int next_int(char *s)
{
  int rc = 0;
  int x,y;

  for (x=0; s[x] && !isdigit(s[x]); x++) ;
  if (s[x]) {
    while (isdigit(s[x])) {
      rc *= 10;
      rc += (s[x]-'0');
      x++;
    }
  }
  for (y=0; !y || s[y-1]; y++) s[y] = s[x+y];
  return rc;
}

boolean digit_in(char *s)
{
  int x;

  if (!s) return false;
  for (x=0; s[x]; x++)
    if (isdigit(s[x]))
      return true;
  return false;
}


void display_current(clauselist clauses)
{
  clauselist cptr;
  int x;
  boolean ss = false;
  boolean ns = false;
  boolean ls = false;

  system("clear");
  for (x=1,cptr=clauses; cptr; cptr=cptr->next) {
    if (cptr->c->sacred) {
      if (too_long_to_be_soft(cptr->c->weight,cptr->c->var_count))
	ls = true;
      else ss = true;
    }
    else ns = true;
  }
  if (ls) {
   printf("Too long - required fixed as true:\n");
    for (x=1,cptr=clauses; cptr; cptr=cptr->next)
      if (cptr->c->sacred &&
	  too_long_to_be_soft(cptr->c->weight,cptr->c->var_count))
	printf("%4d.%s\n", x++, cptr->c->s.c_str);
    printf("\n");
  }
  if (ss) {
    printf("Fixed as true:\n");
    for (x=1,cptr=clauses; cptr; cptr=cptr->next)
      if (cptr->c->sacred &&
	  !too_long_to_be_soft(cptr->c->weight,cptr->c->var_count))
	printf("%4d.%s\n", x++, cptr->c->s.c_str);
    printf("\n");
  }
  if (ns) {
    printf("Not fixed:\n");
    for (x=1,cptr=clauses; cptr; cptr=cptr->next)
      if (!cptr->c->sacred)
	printf("%4d.%s\n", x++, cptr->c->s.c_str);
    printf("\n");
  }
}



dialogue new_dialogue(dialogue d, event e)
{
  int x;
  dialogue newd;

  newd = (dialogue) malloc(sizeof(DIALOGUE));
  newd->parent = d;
  newd->e = e;
  newd->parameters = 0;
  newd->time = theTimelimit;
  for (x=0; x<MAX_SUB_DIALOGUE; x++) newd->sub[x] = 0;
  if (d) {
    for (x=0; d->sub[x]; x++) ;
    d->sub[x] = newd;
  }
  return newd;
}



/*****************************************************************/

void old_generate_diagnosis(char *sorts_and_functions, clauselist clauses)
{
  char tempfilename[16];
  FILE *basis;
  FILE *snf;
  char line[1000];
  char command[1000];
  clauselist cptr;
  int good;
  dynstring clause_input;
  dynstring setting_input;

  snf = fopen(sorts_and_functions,"r");
  if (!snf) inerror("Failed to open input file",sorts_and_functions);
  nonce_filename(tempfilename);
  basis = fopen(tempfilename,"w");
  if (!basis) inerror("Failed to open temporary file",tempfilename);
  while (fgets(line,1000,snf)) fputs(line,basis);
  fclose(snf);
  for (cptr=clauses; cptr; cptr=cptr->next)
    if (cptr->c->sacred)
      fprintf(basis,"clause { %s }\n", cptr->c->s.c_str);
  fprintf(basis,"setting verbosity job none\n");
  fprintf(basis,"setting verbosity models none\n");
  fprintf(basis,"setting verbosity stats none\n");
  fprintf(basis,"setting time_limit 10\n");
  fprintf(basis,"setting solutions 1\n");
  fprintf(basis,"setting random 1\n");
  fprintf(basis,"setting no_print_headersq\n");
  fprintf(basis,"end\n");
  fclose(basis);
  good = init_finder(tempfilename);
  if (!good)
    report_no_model(clauses);
  else {
    dynstring_initialise(&setting_input,"setting time_limit 1");
    finder_input(&setting_input);
    for (cptr=clauses; cptr; cptr=cptr->next)
      if (cptr->c->sacred)
	cptr->c->satisfied = true;
      else {
	dynstring_initialise(&clause_input,"clause { ");
	dynstring_strcat(&clause_input,&(cptr->c->s));
	dynstring_strcatc(&clause_input," }");
	cptr->c->satisfied = found(&clause_input,1,1);
      }
    /*
     * Soft constraints could go here to improve the model.
     * Alternatively, each added clause could be soft when added, and
     * made hard if modelled completely. This requires a facility to
     * retract clauses from the FINDER database, or a switch on found.
     */
    for (cptr=clauses; cptr; cptr=cptr->next)
      if (!cptr->c->satisfied) {
	dynstring_initialise(&clause_input,"clause { ");
	dynstring_strcat(&clause_input,&(cptr->c->s));
	dynstring_strcatc(&clause_input," }");
	cptr->c->satisfied = found(&clause_input,1,0);
      }
    report_model(clauses);
  }
  sprintf(command,"rm -f %s",tempfilename);
  system(command);
  FINDER_cleanup();
}
