#include "FINDER.h"

#ifdef SCOTT
#include "SFshared_protos.h" // protos shared by SCOTT and FINDER
#endif

/*
*	These two are as trivial as they come.
*/

void print_the_job(void)
{
  fprint_the_job(stdout);
}

void print_job(void)
{
  fprint_job(stdout);
}




/*
*	This is the same as fprint_job except that the job is printed in full
*	regardless of the verbosity settings. It is not called from within
*	FINDER.
*/

void fprint_the_job(FILE *fil)
{
  boolean tmp;

  tmp = setting.pm.verbosity[pjob];
  setting.pm.verbosity[pjob] = full;
  fprint_job(fil);
  setting.pm.verbosity[pjob] = tmp;
}



/*
*	Send the job description to fil, in FINDER input format.
*/

void fprint_job(FILE *fil)
{
  if (!setting.pm.verbosity[pjob]) return;
  print_sorts(fil);
  print_functions(fil);
  if (setting.pm.verbosity[pjob] == full) {
    print_clauses(fil);
    print_setting(fil);
  }
  fprintf(fil,"\nend\n");
}



/*
*	First obvious subroutine of fprint_job.
*/

/*
void print_sorts(FILE *fil)
{
  int i;

  fprintf(fil,"sort {\n");
  for (i = 0; sort[i].sname; i++) {
    if (i) fprintf(fil,"\n");
    print_one_sort(fil,sort+i);
  }
  fprintf(fil,"}\n");
}
*/

void print_sorts(FILE *fil)
{
  int i;

  for (i = 0; sort[i].sname; i++) {
    fprintf(fil,"sort ");
    print_one_sort(fil,sort+i);
  }
}


/*
*	Exactly what it says.
*/

void print_one_sort(FILE *fil, kind k)
{
  name_list ep;

  if (k->piped) fprintf(fil," %s { %c piped  ", k->sname, '%');
  else fprintf(fil," %s { ", k->sname);
  if (k->canonical) {
    fprintf(fil,"enum: ");
    for (ep=k->canonical; ep; ep=ep->nextname) {
      if (ep != k->canonical) fprintf(fil,", ");
      fprintf(fil,ep->name);
    }
    fprintf(fil,".");
  }
  else if (k->sizmin == k->sizmax)
    fprintf(fil,"cardinality = %d", k->sizmin);
  else fprintf(fil,"cardinality > %d  cardinality < %d",
	       k->sizmin-1, k->sizmax+1);
  fprintf(fil," }\n");
}



/*
*	Second obvious subroutine of fprint_job.
*/

void print_functions(FILE *fil)
{
  functor f;

  for (f = FN; f; f = f->nextfunct)
    if (setting.pm.pretty || f->pmode) {
      fprintf(fil,"function ");
      print_one_function(fil, f);
    }
}


/*
*	Describing a function symbol is more complex than describing a sort,
*	but conceptually it is no deeper.
*/

void print_one_function(FILE *fil, functor f)
{
  int i;

  fprintf(fil,"%s: ", f->fname);
  for (i=0; i<f->adicity; i++) {
    if (i) fprintf(fil,", ");
    if (f->arg_type[i])
      fprintf(fil,f->arg_type[i]->sname);
    else fprintf(fil,"??");
  }
  if (f->adicity) fprintf(fil," -> ");
  fprintf(fil,"%s", f->val_type->sname);
  if (!f->adicity) fprintf(fil,".");
  fprintf(fil," {\n");
  if (f->total) fprintf(fil,"  total\n");
  else fprintf(fil,"  partial\n");
  if (f->injective && f->surjective) fprintf(fil,"  bijective\n");
  else {
    if (f->injective) fprintf(fil,"  injective\n");
    if (f->surjective) fprintf(fil,"  surjective\n");
  }	
  if (f->leftin && f->leftsur) fprintf(fil,"  column_bijective\n");
  else {
    if (f->leftin) fprintf(fil,"  column_injective\n");
    if (f->leftsur) fprintf(fil,"  column_surjective\n");
  }
  if (f->rightin && f->rightsur) fprintf(fil,"  row_bijective\n");
  else {
    if (f->rightin) fprintf(fil,"  row_injective\n");
    if (f->rightsur) fprintf(fil,"  row_surjective\n");
  }
  if (f->comm) fprintf(fil,"  commutative\n");
  if (f->cut_it) fprintf(fil,"  cut\n");
  if (f->nop) fprintf(fil,"  no_priority\n");
  if (!f->labels) fprintf(fil,"  no_labels\n");
  fprintf(fil,"  scope: %d\n", f->scope);
  if (f->pmode == null) fprintf(fil,"  hidden\n");
  else if (f->pmode == brief) 
    fprintf(fil,"  print: brief\nt");
  else	fprintf(fil,"  print: full\n");
  fprintf(fil,"}\n");
}


/*
*	Third obvious subroutine of fprint_job.
*/

void print_clauses(FILE *fil)
{
  clause c;

  for (c=TH[current_theory]; c; c=c->nextclause) {
  print_one_clause(fil,c);
  }
}


/*
*	Printing a clause is simple. Add the weight, softweight and position.
*/

void print_one_clause(FILE *fil, clause c)
{
  wff_stack wf;

  if (!fil) fil = stdout;
  fprintf(fil,"clause (%d) { ",c->softweight);
  for (wf=c->prem; wf; wf=wf->nxt) {
    if (wf != c->prem) fprintf(fil,", ");
    outwff(fil,wf->w);
  }
  if (c->prem) fprintf(fil,"  ->  ");
  for (wf=c->conc; wf; wf=wf->nxt) {
    if (wf != c->conc) fprintf(fil,"; ");
    outwff(fil,wf->w);
  }
  if (!c->conc) fprintf(fil,"FALSE");
  fprintf(fil,". }");
  fprintf(fil,"  %c %d %d %d %d ",
	  '%', c->varcount, c->weight, c->softweight, 
	  c-theory[current_theory]);
  /*
   * Print variables and their sorts for reference
   */
  for (wf=c->varlist; wf; wf=wf->nxt)
    fprintf(fil," %s %s",wf->w->sym->fname,wf->w->sym->val_type->sname);
  fprintf(fil,"\n");
}



/*
*	Recursively print a formula, in Polish except for =, <, >, +. -.
*/

void outwff(FILE *fil, wff w)
{
  int i;

  if (!w) return;
  if (!fil) fil = stdout;
  if (strchr("=><+-",w->sym->fname[0])) {
    fprintf(fil,"(");
    outwff(fil,w->sb[0]);
    fprintf(fil," %c ",w->sym->fname[0]);
    outwff(fil,w->sb[1]);
    fprintf(fil,")");
  }
  else {
    fprintf(fil,"%s",w->sym->fname);
    if (w->sym->adicity) {
      fprintf(fil,"(");
      for (i=0; i<w->sym->adicity; i++) {
	if (i) fprintf(fil,",");
	outwff(fil,w->sb[i]);
      }
      fprintf(fil,")");
    }
  }
  fflush(fil);
}




/*
*	Fourth obvious subroutine of fprint_job.
*/

void print_setting(FILE *fil)
{
    fprintf(fil,"\nsetting {\n");
    prints_int(fil, "solutions", setting.max_models);
    prints_int(fil, "backtracks", setting.max_backtracks);
    prints_int(fil, "ref_length", setting.maxref);
    prints_int(fil, "test_ref_length", setting.test_maxref);
    prints_int(fil, "time_limit", setting.maxtime);
    prints_int(fil, "pre_test", setting.pretest_limit);
    prints_bool(fil, "primary", "secondary", setting.Nosecs);
    prints_bool(fil, "reverse_test", "obverse_test", setting.invert_change);
    prints_bool(fil, "pretty", "ugly", setting.pm.pretty);
    prints_bool(fil, "noisy", "quiet", setting.noisy);
    prints_bool(fil, "time_stamp", "no_time_stamp", setting.timestamps);
    prints_bool(fil, "dump_depth", "no_dump_depth", setting.dumpdepth);
    prints_bool(fil, "tops_off", "tops_on", setting.topsoff);
    prints_bool(fil, "report_breaks", "silent_breaks", setting.report_breaks);
    prints_bool(fil, "test_by_vars", "test_by_weight", setting.vars_over_wt);
    fprintf(fil,"\tstack: ");
    if (setting.batches == S_ITSY) fprintf(fil,"itsy\n");
    if (setting.batches == S_BITSY) fprintf(fil,"bitsy\n");
    if (setting.batches == S_TINY) fprintf(fil,"tiny\n");
    else if (setting.batches == S_SMALL) fprintf(fil,"small\n");
    else if (setting.batches == S_MEDIUM) fprintf(fil,"medium\n");
    else if (setting.batches == S_LARGE) fprintf(fil,"large\n");
    else if (setting.batches == S_VERYLARGE) fprintf(fil,"LARGE\n");
    else if (setting.batches == S_HUGE) fprintf(fil,"huge\n");
    else if (setting.batches == S_MAXIMAL) fprintf(fil,"maximal\n");
    else fprintf(fil,"%d\n", setting.batches);
    fprintf(fil,"\tverbosity {\n\t");
    prints_verb(fil, "models", pmodels);
    prints_verb(fil, "job", pjob);
    prints_verb(fil, "stats", pstats);
    fprintf(fil,"}\n}\n");
}




/*
*	There now follow the subroutines called by print_set.
*/

void prints_int(FILE *fil, char *s, int x)
{
    if (x) fprintf(fil,"\t%s %d\n", s, x);
}


void prints_bool(FILE *fil, char *strue, char *sfalse, boolean b)
{
    fprintf(fil,"\t%s\n", b? strue: sfalse);
}


void prints_verb(FILE *fil, char *s, print_mode pmd)
{
  prints p;

  p = setting.pm.verbosity[pmd];
  switch(p) {
  case null:
    fprintf(fil,"\t%s: none\n\t", s);
    return;
  case brief:
    fprintf(fil,"\t%s: brief\n\t", s);
    return;
  case full:
    fprintf(fil,"\t%s: full\n\t", s);
    return;
  }
  skipout("Strange verbosity setting",1);
}


/*
=============================================================================
*/


/*
*	Print_model is called when there is a new model to report. It checks
*	for error conditions such as reporting the same model twice and 
*	as a side effect maintains the vector last_good. It also updates the
*	global counter good.
*
*	There are two formats, pretty and ugly (human-readable and machine-
*	readable). In addition, there are two print modes. Brief means print
*	only when there is a change to report, and Full means print no matter 
*	what. There is also None, which is self-explanatory. There is a 
*	general verbosity setting and a local one for each function. The 
*	lower of these two over-rides the higher.
*
*	Before the real thing starts, there are some trivia as in the 
*	above job-printing case.
*/

boolean print_the_model(void)
{
    return fprint_the_model(stdout);
}

boolean fprint_the_model(FILE *fil)
{
    prints tmp;
    boolean rc;

    tmp = setting.pm.verbosity[pmodels];
    setting.pm.verbosity[pmodels] = full;
    rc = fprint_model(fil);
    setting.pm.verbosity[pmodels] = tmp;
    return rc;
}

boolean print_model(void)
{
    return fprint_model(stdout);
}

boolean fprint_model(FILE *fil)
{
  kind k;
  functor f;
  int timenow, num_theories;
  boolean Full, Brief, header;
  int theArguments[ARGMAX];
  int i;

#ifdef SCOTT
  num_theories = get_num_maxisets();
#else
  num_theories = 1;
#endif

  header = setting.print_headers;
  if (setting.pm.verbosity[pmodels] && !setting.pm.pretty)
    fprintf(fil,"model\n");

  if (firstmodel) {
    firstmodel = false;
    if (setting.pm.verbosity[pmodels] && !setting.pm.pretty) {
      for (k = sort+1; k->sname; k++)
	fprintf(fil,"size %s %d\n", k->sname, k->siz);
    }
  }

  if (setting.noisy) fprintf(WARNING_STREAM,"\007");
  if (!(setting.pm.verbosity[pmodels])) return true;
  for (f = FN; f; f = f->nextfunct) 
    if (!negative || f->piped) {

      Full = (f->pmode==full && setting.pm.verbosity[pmodels]==full);

      Brief = ((f->pmode==full && setting.pm.verbosity[pmodels]==brief)
	       || (f->pmode==brief && setting.pm.verbosity[pmodels]));

      if (Full || (Brief && (f->piped || f->fcc <= first_change))) {

	if (header && setting.pm.pretty) {
	  if (num_theories > 1) {
	    if (setting.timestamps) {
	      timenow = timestamp()-begin_time;
	      if (theBadness)
		fprintf(fil,
			"\nModel %d.%d [ %d %d %d ] (%3.2f seconds, %d Inserts)\n",
			current_theory,
			updates[current_theory],
			theBadness,
                        theTime-begin_time,
                        theWork,
			(timenow*1.0)/theClockTick,
			workNow);
	      else
		fprintf(fil,
			"\nModel %d.%d (%3.2f seconds, %d Inserts)\n",
			current_theory,
			updates[current_theory],
			(timenow*1.0)/theClockTick,
			workNow);
	    }
	    else {
	      if (theBadness)
		fprintf(fil,"\nModel %d.%d [ %d %d %d ] (%d Inserts)\n",
			current_theory,
			updates[current_theory],
			theBadness,
			theTime-begin_time,
			theWork,
			workNow);
	      else
		fprintf(fil,"\nModel %d.%d\n",
			current_theory,
			updates[current_theory]);
	    }
	  }
	  else {
	    if (setting.timestamps) {
	      timenow = timestamp()-begin_time;
	      if (theBadness)
		fprintf(fil,
			"\nModel %d [ %d %d %d ] (%3.2f seconds, %d Inserts)\n",
			updates[0],
			theBadness,
			theTime-begin_time,
			theWork,
			(timenow*1.0)/theClockTick,
			workNow);
	      else
		fprintf(fil,
			"\nModel %d (%3.2f seconds, %d Inserts)\n",
			updates[0],
			(timenow*1.0)/theClockTick,
			workNow);
	    }
	    else {
	      if (theBadness)
		fprintf(fil,"\nModel %d [ %d %d %d ] (%d Inserts)\n",
			updates[0],
			theBadness,
			theTime-begin_time,
			theWork,
			workNow);
	      else
		fprintf(fil,"\nModel %d\n",
			updates[0]);
	    }
	  }
	  header = false;
	}

	for (i=0; i<ARGMAX; i++) theArguments[i] = 0;

	if (f->adicity > 1)
	  multiprint(fil,f,0,theArguments);
	else switch(f->adicity) {
	case 0:
	  zeroprint(fil, f);
	  break;
	case 1:
	  oneprint(fil, f);
	  break;
	}
	fprintf(fil,"\n");
      }
    }
  if (!setting.pm.pretty) fprintf(fil,".\n");
  else if (!header) fprintf(fil,"\n");
  return true;
}



/*
*	Print a zero-adic function.
*/

void zeroprint(FILE *fil, functor f)
{
    if (setting.pm.pretty) {
	fprintf(fil,"\n%s = ", f->fname);
	printvalue(fil, f->val_type, *(f->table), 0);
    }
    else fprintf(fil,"%s %d", f->fname, *(f->table));
}


/*
*	Print a monadic function.
*/

void oneprint(FILE *fil, functor f)
{
  int indentation;
  int *columnwidth = 0;       /* Array */
  int i, j;

  if (setting.pm.pretty) {
    indentation = get_monadic_indentation(f);
    columnwidth = get_monadic_columnwidth(f);

    if (f->labels) {
      fprintf(fil,"\n");
      indent(fil,indentation-strlen(f->arg_type[0]->sname));
      fprintf(fil,"%s | ", f->arg_type[0]->sname);
      for (i = 0; i < f->arg_type[0]->siz; i++)
	printvalue(fil, f->arg_type[0], i, columnwidth[i]);
      fprintf(fil,"\n");
      for (i = 0; i <= indentation; i++) fprintf(fil,"-");
      fprintf(fil,"+-");
      for (j = 0; j < f->arg_type[0]->siz; j++) {
	for (i = 0; i < columnwidth[j]-1; i++)
	  fprintf(fil,"-");
	if (j < f->arg_type[0]->siz-1)
	  fprintf(fil,"--");
      }
      fprintf(fil,"\n");
      indent(fil,indentation-strlen(f->fname));
      fprintf(fil,"%s | ", f->fname);
    }
    for (i = 0; i < f->arg_type[0]->siz; i++)
      printvalue(fil, f->val_type, f->table[i+1], columnwidth[i]);
  }
  else {
    fprintf(fil,"%s ", f->fname);
    for (i = 0; i < f->arg_type[0]->siz; i++)
      fprintf(fil," %d", *(f->table+i+1));
  }
  if (columnwidth) free(columnwidth);
}



/*
 * Print a function of adicity greater than 1
 */

void multiprint(FILE *fil, functor f, int x, int theArguments[])
{
  int v;
  dynstring stuff;
  int indentation = 0;
  int *columnwidth = 0;       /* Array */
  int i,j;

  dynstring_initialise(&stuff, "");

  if (!setting.pm.pretty) {
    if (!x) fprintf(fil,"%s ", f->fname);
    for (theArguments[x] = 0;
	 theArguments[x] < f->arg_type[x]->siz;
	 theArguments[x]++) {
      if (x == f->adicity-1)
	fprintf(fil," %d",*(f->table + getoffset(f,theArguments)));
      else multiprint(fil,f,x+1,theArguments);
    }
  }
  else {
    if (x < f->adicity-2) {
      for (theArguments[x]=0;
	   theArguments[x] < f->arg_type[x]->siz;
	   theArguments[x]++)
	multiprint(fil,f,x+1,theArguments);
    }
    else {
      columnwidth = get_column_width(f,theArguments);
      if (f->labels) {
	indentation = get_indent(f,theArguments);
	fprintf(fil,"\n");
	if (f->adicity == 2) {
	  indent(fil,indentation - (strlen(f->fname)+1));
	  fprintf(fil, "%s ", f->fname);
	}
	else {
	  i = strlen(f->fname)+5;
	  for (v=0; v<f->adicity-2; v++) {
	    if (v) i += 1;
	    i += f->arg_type[v]->len[theArguments[v]];
	  }
	  indent(fil,indentation-i);
	  fprintf(fil, "%s(", f->fname);
	  for (v=0; v<f->adicity-2; v++) {
	    if (v) fprintf(fil,",");
	    get_the_name(f->arg_type[v],theArguments[v],&stuff);
	    fprintf(fil, "%s", stuff.c_str);
	  }
	  fprintf(fil,",_) ");
	}
	fprintf(fil,"|");
	for (i=0; i < f->arg_type[x+1]->siz; i++) {
	  get_the_name(f->arg_type[x+1],i,&stuff);
	  fprintf(fil," %s",stuff.c_str);
	  indent(fil,columnwidth[i]- stuff.length);
	}
	fprintf(fil,"\n");
	for (i=0; i<indentation; i++) fprintf(fil,"-");
	fprintf(fil,"+");
	for (i=0; i < f->arg_type[x+1]->siz; i++)
	  for (j=0; j<=columnwidth[i]; j++)
	    fprintf(fil,"-");
	fprintf(fil,"\n");
      }
      for (theArguments[x] = 0;
	   theArguments[x] < f->arg_type[x]->siz;
	   theArguments[x]++) {
	if (f->labels) {
	  get_the_name(f->arg_type[x],theArguments[x],&stuff);
	  indent(fil,indentation-(stuff.length+1));
	  fprintf(fil,"%s |",stuff.c_str);
	}
	else indent(fil,indentation+1);
	for (theArguments[x+1] = 0;
	     theArguments[x+1] < f->arg_type[x+1]->siz;
	     theArguments[x+1]++) {
	  get_the_value(f,theArguments,&stuff);
	  fprintf(fil," %s",stuff.c_str);
	  if (theArguments[x+1] < f->arg_type[x+1]->siz)
	    indent(fil,columnwidth[theArguments[x+1]]-stuff.length);
	}
	fprintf(fil,"\n");
      }
    }
  }

  if (columnwidth) free(columnwidth);
  dynstring_free(&stuff);
}


void indent(FILE *fil, int x)
{
  int i;

  for (i=0; i<x; i++) fprintf(fil," ");
}



int get_monadic_indentation(functor f)
{
  int len1, len2;

  len1 = strlen(f->arg_type[0]->sname);
  len2 = strlen(f->fname);
  if (len1 > len2) return len1;
  return len2;
}


int *get_monadic_columnwidth(functor f)
{
  int *cw;
  int x;
  int siz = f->arg_type[0]->siz;

  cw = (int*) malloc(siz * sizeof(int));
  for (x=0; x<siz; x++) {
    cw[x] = 2;
    if (f->arg_type[0]->len[x] > cw[x])
      cw[x] = f->arg_type[0]->len[x];
    if (f->table[x+1] < f->val_type->siz &&
	f->val_type->len[f->table[x+1]] > cw[x])
      cw[x] = f->val_type->len[f->table[x+1]];
    cw[x]++;
  }
  return cw;
}


int get_indent(functor f, int theArguments[])
{
  int oldlen;
  int newlen;
  int v;

  oldlen = 2;

  newlen = strlen(f->fname)+1;
  if (f->adicity > 2) newlen += 4;
  for (v=0; v<f->adicity-2; v++) {
    if (v) newlen += 1;
    newlen += f->arg_type[v]->len[theArguments[v]];
  }
  if (newlen > oldlen) oldlen = newlen;

  for (v=0; v < f->arg_type[f->adicity-2]->siz; v++)
    if (f->arg_type[f->adicity-2]->len[v] >= oldlen)
      oldlen = f->arg_type[f->adicity-2]->len[v]+1;
  return oldlen;
}



int *get_column_width(functor f, int theArguments[])
{
  int i,j;
  dynstring stuff;
  int *cw = 0;

  dynstring_initialise(&stuff, "");

  if (f->adicity > 0) {
    i = f->adicity-1;
    cw = (int*) malloc(f->arg_type[i]->siz * sizeof(int));
    for (j=0; j < f->arg_type[i]->siz; j++) cw[j] = 2;
    for (j=0; j < f->arg_type[i]->siz; j++)
      if (f->arg_type[i]->len[j] > cw[j])
	cw[j] = f->arg_type[i]->len[j];
    if (f->adicity > 1) {
      for (theArguments[i-1] = 0;
	   theArguments[i-1] < f->arg_type[i-1]->siz;
	   theArguments[i-1]++) {
	for (theArguments[i] = 0;
	     theArguments[i] < f->arg_type[i]->siz;
	     theArguments[i]++) {
	  get_the_value(f,theArguments,&stuff);
	  j = stuff.length;
	  if (j > cw[theArguments[i]])
	    cw[theArguments[i]] = j;
	}
      }
    }
    else {
      for (theArguments[0] = 0;
	   theArguments[0] < f->arg_type[0]->siz;
	   theArguments[0]++) {
	get_the_value(f,theArguments,&stuff);
	  j = stuff.length;
	  if (j > cw[theArguments[i]])
	    cw[theArguments[i]] = j;
      }
    }
  }

  dynstring_free(&stuff);
  return cw;
}



void get_the_value(functor f, int theArguments[], dynstring *dest)
{
  get_the_name(f->val_type,
	       *(f->table + getoffset(f,theArguments)),
	       dest);
}


void get_the_name(kind k, int x, dynstring *dest)
{
  name_list lp;
  int i;

  if (x >= k->siz) dynstring_strcpyc(dest," ");
  else if (k->canonical) {
    lp = k->canonical;
    for (i=0; i<x; i++) lp = lp->nextname;
    dynstring_strcpyc(dest,lp->name);
  }
  else
    dynstring_sprintf(dest,"%d",x);
}



/*
*	Return the length of the longest canonical name of sort k.
*/

int canonlength(kind k)
{
  int i = 0, j;

  for (j = 0; j < k->siz; j++)
    if (k->len[j] > i) i = k->len[j];
  return i;
}




/*
*	Print the canonical name of object x of kind k, padded on the right 
*	with spaces up to the length len.
*/

void printvalue(FILE *fil, kind k, int x, int len)
{
  name_list lp;
  int i, j;

  if (x == k->siz) j = 0;
  else if (k->canonical && setting.pm.pretty) {
    lp = k->canonical;
    for (i = 0; i < x; i++) lp = lp->nextname;
    fprintf(fil,lp->name);
    j = strlen(lp->name);
  }
  else {
    j = 2;
    fprintf(fil,"%-2d",x);
  }
  for (i = j; i <= len; i++) fprintf(fil," ");
  fflush(stdout);
}




/*
*	Rather similar, for printing the arguments of the function.
*/

void printarg(FILE *fil, kind k, int x, int len)
{
  name_list lp = 0;
  int i, j;

  if (x == k->siz) j = 0;
  else if (k->canonical) {
    lp = k->canonical;
    for (i = 0; i < x; i++) lp = lp->nextname;
    j = strlen(lp->name);
  }
  else j = 2;
  for (i = j; i < len; i++) fprintf(fil," ");

  if (x == k->siz) return;
  if (k->canonical) fprintf(fil,lp->name);
  else fprintf(fil,"%2d",x);
  fflush(stdout);
}



/*
============================================================================
*/


/*
*	When the FINDER job has been completed, some statistics on the run
*	are printed out. These are determined by the search controller used.
*/

void printstats(void)
{
  void dump_report();

  printf("\n\n\n");
  dump_report(&tr_parameters);
  printf("\n");
  dump_report(&ttr_par);
  printf("FINDER time: %3.2f seconds\n\n",
	 ((end_time-begin_time)*1.0) / theClockTick);
}



void dump_report(trin tr)
{
  int i;
  logical t;

  for (i=0; i<MAX_REPORT && (t = tr->reps[i].report_type); i++) {
    if (i == SOLUTIONS)
      printf("%s %d\n", tr->reps[i].s_report, tr->reps[i].i_report);
    else if ((t & INT_REPORT) && tr->reps[i].i_report)
      printf("%s %d\n", tr->reps[i].s_report, tr->reps[i].i_report);
    if ((t & FLOAT_REPORT) && tr->reps[i].f_report)
      printf("%s %3.2f\n",tr->reps[i].s_report,tr->reps[i].f_report);
  }
  for (i=0; i<MAX_REPORT && (t = tr->reps[i].report_type); i++)
    if (t & DO_REPORT)
      (*tr->reps[i].do_report)(tr);
}



void dump_trace(trin tr, boolean final)
{
  int i;
  logical t;
  int x = 0;

  tr->tracepoints++;
  if (final || !(tr->tracepoints % tr->dump_rate)) {
    if (final) printf("-1\n");
    for (i=0; i<MAX_REPORT; i++) {
      t = tr->reps[i].report_type;
      if (t & INT_REPORT) {
	if (final) printf("%d. %s (int)\n", ++x, tr->reps[i].s_report);
	else printf(" %d", tr->reps[i].i_report);
      }
      if (t & FLOAT_REPORT) {
	if (final) printf("%d. %s (float)\n", ++x, tr->reps[i].s_report);
	else printf(" %0.2f", tr->reps[i].f_report);
      }
    }
    printf("\n");
  }
}



void ref_size_report(void)
{
  int i, j, k, m;
  int ctotal[RRB], rtotal[RRA], totaltotal;

  if (0) {
    totaltotal = 0;
    for (i=0; i<RRB; i++) {
      ctotal[i] = 0;
      for (j=0; j<RRA; j++)
	ctotal[i] += tr_parameters.ref_record[j][i];
    }
    for (i=0; i<RRA; i++) {
      rtotal[i] = 0;
      for (j=0; j<RRB; j++)
	rtotal[i] += tr_parameters.ref_record[i][j];
      totaltotal += rtotal[i];
    }

    for (k=RRA-1; k >= 0; k--) {
      j = 0;
      for (i=0; i<RRB; i++)
	if (tr_parameters.ref_record[k][i]) j = 1;
      if (j) break;
    }
    if (k < 0) return;
    for (m=RRB-1; m >= 0; m--) {
      j = 0;
      for (i=0; i <= k; i++)
	if (tr_parameters.ref_record[i][m]) j = 1;
      if (j) break;
    }
    if (!m) return;
    dump_refs(k+1,m+1);
  }
}




void report_branches(trin tr)
{
  int i, tsz = 0;

  if (tr->branches) {
    for (tsz=tr->maxval; tsz && !tr->branches[tsz]; tsz--) ;
    if (tsz) {
      printf("Branching: [%d]",tr->branches[1]);
      for (i=2; i <= tsz; i++)
	printf(" %d",tr->branches[i]);
      printf("\n");
    }
  }
}


void stopped_by(trin tr)
{
  if (tr->reps[STOP_CODE].s_report[0])
    printf("Stopped by %s\n",tr->reps[STOP_CODE].s_report);
  else
    printf("Stopped by exhaustion\n");
}




/*
*	The distribution matrix of the refutations is dumped to file
*	FINDER.refs. First the numbers of rows and columns are printed
*	along with the maximum number in the grid.
*/

void dump_refs(int rows, int columns)
{
  FILE *ref_file;
  int i, j, ccmax = 0;

  for (i = rows-1; i >= 0; i--)
    for (j = 0; j < columns; j++)
      if (tr_parameters.ref_record[i][j] > ccmax)
	ccmax = tr_parameters.ref_record[i][j];
  if (!(ref_file = fopen("FINDER.refs","w"))) return;
  fprintf(ref_file,"%d %d %d\n", rows, columns, ccmax);

  for (i = rows-1; i >= 0; i--) {
    for (j = 0; j < columns; j++)
      fprintf(ref_file, " %d", tr_parameters.ref_record[i][j]);
    fprintf(ref_file, "\n");
  }
  fclose(ref_file);
}


/*
 * The next two functions print the constraint graph in the riht
 * format to be used by CGV.
 */

void print_constraint_graph(trs T)
{
  int x, y;
  reflist r;

  if (!(T->job->profile)) return;

  printf("g12vis start\n");
  /* Report the variables */
  for (x=0; x<T->job->vlength; x++)
    if (T->vector[x].poss > 1) {
      printf("add var v%d => ", x);
      print_un_vec_pos(x);
      printf("\n");
    }

  /* Report the constraints */
  for (x=0; x<T->job->vlength; x++) {
    for (y=0; y<T->vector[x].csiz; y++) {
      for (r = T->vector[x].cvals[y].cvrefs; r; r = r->nextref)
        check_new_ref(x, r->rf, T);
      for (r = T->vector[x].cvals[y].posrefs; r; r = r->nextref)
        check_new_ref(x, r->rf, T);
    }
    for (r = T->vector[x].inject; r; r = r->nextref)
      check_new_ref(x, r->rf, T);
  }
  printf("g12vis end\n");
}



void check_new_ref(int x, ref rf, trs T)
{
  intlist il;
  int count = 0;

  if (!rf) return;
  for (il=rf->cellvals; il; il=il->p) {
    if (il->v < x)
      return;
    if (T->vector[il->v].cvals[il->w].blocks && !(il->p))
      return;
    if (T->vector[il->v].poss == 1) {
      if (!(T->vector[il->v].cvals[il->w].blocks) && il->p)
        return;
    }
    else
      count++;
  }

  if (count>1) {
    printf("add con c%d [", (int)rf);
    for (il=rf->cellvals; il; il=il->p)
      if (T->vector[il->v].poss > 1)
        printf(" v%d", il->v);
    printf(" ]\n");
  }
}


/*
 * For now, dump the domains. Later maybe dump a CGV batch of changes
 * since the last report
 */

void cgv_state_report(trs T, cell c, int val)
{
  set *domains;
  int x,y;
  functor f;
  int args[ARGMAX];

  if (T->job->vlength)
    domains = (set*) malloc(T->job->vlength * sizeof(set));
  for (x=0; x<T->job->vlength; x++) {
    domains[x] = emptyset(0);
    if (T->vector[x].has_value)
      intoset(T->icomvec[x],domains[x]);
    else
      for (y=0; y<T->vector[x].csiz; y++)
	if (!(T->vector[x].cvals[y].blocks))
	  intoset(y,domains[x]);
  }
  printf("cgv start\n");
  dump_state(domains,T);
  if (c) {
    printf("cgv assigned ");
    f = un_vec_pos(c->vpos,args);
    printf("%s",f->fname);
    for (x=0; x<f->adicity; x++) printf(" %d",args[x]);
    printf(" %d\n",val);
  }
  else printf("cgv solution\n");
  printf("cgv end\n");
  for (x=0; x<T->job->vlength; x++)
    freeset(domains[x]);
  if (domains) free(domains);
}


void dump_state(set domains[], trs T)
{
  functor f;
  int argument[ARGMAX];
  int x,y,z;
  boolean flag = false;
  boolean real_value;

  for (f=FN; f; f=f->nextfunct)
    if (f->cgv) {
      flag = true;
      printf("cgv %s %d ",f->fname,f->adicity);
      for (x=0; x<f->adicity; x++)
	printf("%d ",f->arg_type[x]->siz);
      for (x=0; x<f->adicity; x++) argument[x] = 0;
      do {
	y = vec_pos(f,argument);
	for (real_value=true,x=0; x<f->adicity; x++)
	  if (argument[x] == f->arg_type[x]->siz) {
	    real_value = false;
	    break;
	  }
	if (real_value || !(f->total))
	  print_cgv_set(domains[y],T->vector[y].csiz);
      }
      while (another_argument(f,argument));
    }
  if (flag) printf("\n");
}


/*
 * Will be more sophisticated if real CGV output is ever required
 */

void print_cgv_set(set domain, int maxval)
{
  int x;

  printf("{");
  for (x=0; x<maxval; x++)
    if (isinset(x,domain))
      printf(" %d",x);
  printf(" } ");
}
