#include "FINDER.h"


static int COMM_SZ = 0;
static int COMM_ARGMAX = 0;
static int **COMM_UPTO = 0;

/*
 * The values of function f are kept in an array v1...vn where n is
 * the number of possible tuples of arguments. v0 is the location of
 * the dummy value taken in the case where arguments are out of range.
 */

int getoffset(functor f, int argument[])
{
  int i;
  int offset = 1;
  int commut_vec[ARGMAX];
  int x;

  if (!f->adicity) return 0;
  if (f->comm) {
    for (i=0; i<f->adicity; i++)
      if (argument[i] >= f->arg_type[i]->siz || argument[i] < 0)
	return 0;
    for (i=0; i<f->adicity; i++)
       commut_vec[i] = argument[i];
    for (i=1; i<f->adicity; i++)     /* Yes I know, but adicity < 4, OK? */
      if (commut_vec[i-1] < commut_vec[i]) {
	x = commut_vec[i];
	commut_vec[i] = commut_vec[i-1];
	commut_vec[i-1] = x;
	if (i > 1) i -= 2; else i-= 1;
      }
    offset = commutative_offset(f->adicity,commut_vec);
  }
  else {
    for (i=0; i<f->adicity; i++) {
      if (f->arg_type[i] &&
	  (argument[i] >= f->arg_type[i]->siz || argument[i] < 0))
	return 0;
      offset += (argument[i] * f->fats[i]);
    }
  }
  return offset;
}


/*
 * The offset in the case of a generalised commutative function,
 * assuming the n arguments have been sorted into non-increasing
 * order. We are enumerating the non-increasing vectors of
 * integers. First come those whose head term is less than argument[0]
 * (we just add the number). On top of that, we enumerate those whose
 * head term is exactly argument[0]. The call is thus recursive.
 */

int commutative_offset(int n, int argument[])
{
  if (n < 1) return 0;
  if (n == 1) return argument[0]+1;
  if (argument[0] == 0) return 1;
  return
    commutative_upto(n,argument[0]-1) +
    commutative_offset(n-1,argument+1);
}


/*
 * The number of non-increasing vectors of n non-negative integers
 * whose leading (head) term is lead or less. This is the number whose
 * head term is exactly lead (= the number of length n-1 whose head is
 * lead or less) plus the number whose head term is lead-1 or less. We
 * dispose of some junk or trivial cases first.
 */

int commutative_upto(int n, int lead)
{
  if (n < 1 || lead < 0) return 0;       /* Impossible cases */
  if (n == 1) return lead+1;             /* <0> ... <lead> */
  if (lead == 0) return 1;               /* <0,...,0> */
  if (n <= COMM_ARGMAX && lead <= COMM_SZ) {
    if (!COMM_UPTO[n][lead])
      COMM_UPTO[n][lead] =
	commutative_upto(n,lead-1) + commutative_upto(n-1,lead);
    return COMM_UPTO[n][lead];
  }
  else
    return commutative_upto(n,lead-1) + commutative_upto(n-1,lead);
}



int vec_pos(functor f, int argument[])
{
  /*
   *  if (f->piped) skipout("Piped function in vector",SKIP);
   */
  return f->fcc + getoffset(f,argument);
}




/*
*	This simply counts the non-constant symbols in the latest
*	clause (merging duplicates) and assigns the result to its 
*	"weight" field. It also counts the constituent variables 
*	for its "varcount".
*/


void get_weight(clause c)
{
  wff_stack wf;
  int wt;
  functor f;
  boolean piped_content = false;

  wt = 0;
  for (wf=c->testlist; wf; wf=wf->nxt) {
    f = wf->w->sym;
    if (f->piped) piped_content = true;	
    if (is_open(f) && f->category != Const)
      wt++;
  }

  if (!(c->weight=wt) && !piped_content)
    inwarning("Clause contains no function symbol",""); 

  c->varcount = 0;
  for (wf=c->varlist; wf; wf=wf->nxt)
    c->varcount++;

#ifndef SOFTIE_SCOTT
  if (c->softweight && too_long_to_be_soft(c->weight,c->varcount)) {
    inwarning("Soft clause is too long","Treat as hard");
    c->softweight = 0;
  }
#endif
}



/*
*	We have just added a new clause to the start of the theory
*	list and set TH to point at it. So move it into the middle
*	of the list until it is in its correct position, then move
*	TH to point back at the start of the list. Note that this 
*	version of normalize assumes that the clauses are added one
*	by one, and that the pre-existing list is already normal.
*/

void normalize(void)
{
  while ( is_after(TH[current_theory], 
		   TH[current_theory]->nextclause) ) {
    TH[current_theory]->nextclause->lastclause = 
      TH[current_theory]->lastclause;

    if ( TH[current_theory]->lastclause )
      TH[current_theory]->lastclause->nextclause = 
	TH[current_theory]->nextclause;

    TH[current_theory]->lastclause = TH[current_theory]->nextclause;

    TH[current_theory]->nextclause = 
      TH[current_theory]->lastclause->nextclause;

    TH[current_theory]->lastclause->nextclause = TH[current_theory];

    if ( TH[current_theory]->nextclause )
      TH[current_theory]->nextclause->lastclause = TH[current_theory];
  }
  while ( TH[current_theory]->lastclause )
    TH[current_theory] = TH[current_theory]->lastclause;
}




/*
*  Clause c1 is after clause c2 if its weight and variable count 
*  say so. This is where the order of clauses is determined. Note
*  that the latest clause is not included.
*
*  A clause of pre-testable weight beats one of greater weight
*  irrespective of any other conditions.
*
*  If both are pre-testable the lighter wins, then the one with 
*  fewer variables, then the accession order.
*
*  If neither is pre-testable the lower weight wins, then the
*  lower variable count (or variables followed by weight if
*  vars_over_wt is set), then accession order.
*/

boolean is_after( clause c1, clause c2 )
{
  if ( !c1 || !c2 ) return false;

  if ( c2->weight < setting.pretest_limit ) {
    if ( c1->weight > c2->weight ) return true;
    if ( c1->weight < c2->weight ) return false;
    if ( c1->varcount > c2->varcount ) return true;
    if ( c1->varcount < c2->varcount ) return false;
    return ( c1 > c2 );
  }
  if ( c1->weight < setting.pretest_limit ) return false;
  
  if ( setting.vars_over_wt ) {
    if ( c1->varcount > c2->varcount ) return true;
    if ( c1->varcount < c2->varcount ) return false;
    if ( c1->weight > c2->weight ) return true;
    if ( c1->weight < c2->weight ) return false;
  }
  else {
    if ( c1->weight > c2->weight ) return true;
    if ( c1->weight < c2->weight ) return false;
    if ( c1->varcount > c2->varcount ) return true;
    if ( c1->varcount < c2->varcount ) return false;
  }
  return ( c1 > c2 );
}



void init_comm_upto(void)
{
  functor f;
  int i,j;

  if (COMM_UPTO) {
    for (i=0; i<COMM_ARGMAX; i++)
      if (COMM_UPTO[i]) free(COMM_UPTO[i]);
    free(COMM_UPTO);
  }
  COMM_ARGMAX = 2;
  COMM_SZ = Integer->siz;
  for (f=FN; f; f=f->nextfunct) if (f->comm) {
      if (f->adicity>COMM_ARGMAX)
	COMM_ARGMAX = f->adicity;
      if (f->val_type->siz > COMM_SZ)
	COMM_SZ = f->val_type->siz;
    }
  COMM_UPTO = (int**) malloc((COMM_ARGMAX + 1) * sizeof(int*));
  for (i=0; i<=COMM_ARGMAX; i++)
    COMM_UPTO[i] = (int*) malloc((COMM_SZ + 1) * sizeof(int));
  for (i=0; i<=COMM_ARGMAX; i++)
    for (j=0; j<=COMM_SZ; j++)
      COMM_UPTO[i][j] = 0;
}

void set_up_cc(void)
{
  functor f;

  init_comm_upto();
  set_sum();
  set_minus();
  set_pred();
  set_succ();
  set_E();
  set_max();
  set_cc();
  set_len();
  tr_parameters.piping = false;
  for (f=FN; f; f=f->nextfunct)
    if (f->piped)
      tr_parameters.piping = true;
  tr_parameters.vlength = 0;
  set_vlength(FN);
  if (tr_parameters.forbidden) free(tr_parameters.forbidden);
  tr_parameters.forbidden =
    (boolean*) malloc(tr_parameters.vlength * sizeof(boolean));
  set_forbidden(FN);
  set_table(FN);

  set_matrix(TH[current_theory]);
  if (latest_clause.prem || latest_clause.conc)
    set_matrix(&latest_clause);

#ifdef DEBUG
  { functor f;
  for ( f = FN; f; f = f->nextfunct ) print_set( f );
  }
#endif

  if ( !(setting.pm.verbosity[pstats]) )
    tr_parameters.verb = 0;
  else if ( setting.pm.verbosity[pstats]==brief )
    tr_parameters.verb = 1;
  else if ( setting.pm.verbosity[pstats]==full )
    tr_parameters.verb = 2;
  tr_parameters.Nosecs = setting.Nosecs;
  tr_parameters.topsoff = setting.topsoff;
  tr_parameters.report_breaks = setting.report_breaks;
  tr_parameters.print_headers = setting.print_headers;
  tr_parameters.random = setting.random;
  tr_parameters.dumpdepth = setting.dumpdepth;
  tr_parameters.dump_rate = setting.dump_rate;
  tr_parameters.profile = setting.profile;
  tr_parameters.awnty_enabled = setting.awnty_enabled;
  tr_parameters.cgv_enabled = setting.cgv_enabled;
  tr_parameters.batches = setting.batches;
  tr_parameters.extra_batches = setting.extra_batches;
  tr_parameters.maxref = setting.maxref;
  tr_parameters.maxbak = setting.max_backtracks;
  tr_parameters.maxinvals = setting.max_invalues;
  tr_parameters.max_penalty = setting.target_bound;
}



/*
 *  This sets the "matrix" field of each formula to point at the table
 *  of values for the corresponding symbol.
 */

void set_matrix(clause c)
{
  if (!c) return;
  set_matrix(c->nextclause);
  set_my_matrix(c->prem);
  set_my_matrix(c->conc);
}

void set_matrices_flat(wff w); // prototype

void set_my_matrix(wff_stack ws)
{
  if (!ws) return;
  set_my_matrix(ws->nxt);
  set_matrices(ws->w);
}


void set_matrices(wff w)
{
  int i;

  if (!w) return;
  for (i=0; i<w->sym->adicity; i++)
    set_matrices( w->sb[i] );
  w->matrix = w->sym->table; 

} /* set_matrices() */



/*
 * The following function mallocs memory for the table of values for
 * each function. The built-in ones for each sort already have their
 * tables allocated at declaration, so they are skipped here.
 * The value returned is "true" (success) or "false" (the required
 * table was too big to fit in the memory available).
 */

void set_table( functor f )
{
  int count, i;

  if (!f) return;
  set_table(f->nextfunct);
  if (f->piped) return;
  if (strchr("=><+-",f->fname[0])) return;
  if (!strcmp("DIF",f->fname) ||
      !strncmp("EST_",f->fname,4) ||
      !strncmp("LEN_",f->fname,4) ||
      !strncmp("PRED_",f->fname,5) ||
      !strncmp("SUCC_",f->fname,5) ||
      !strncmp("MIN_",f->fname,4) ||
      !strncmp("MAX_",f->fname,4) ||
      !strcmp("NOT",f->fname) ||
      !strcmp("AND",f->fname) ||
      !strcmp("OR",f->fname)) return;
  /*
   *  if (f->adicity) count = 2; else count = 1;
   */
  count = 1;                      /* Why allocate twice as much as needed? */
  for (i=0; i<f->adicity; i++)
    count *= (f->arg_type[i]->siz);
  if (f->table) free(f->table);
  f->table = (int*) malloc(sizeof(int) * (count+1));
}



void set_sizmask(void)
{
  int i, k;

  for (k=0; sort[k].siz; k++) {
    if (sort[k].sizmask) free(sort[k].sizmask);
    sort[k].sizmask = emptyset(0);
    for ( i = 0; i < sort[k].siz; i++ )
      intoset(i,sort[k].sizmask);
  }
}


void unset_sizmask(void)
{
  int k;

  for (k=0; sort[k].siz; k++)
    if (sort[k].sizmask) {
      freeset(sort[k].sizmask);
      sort[k].sizmask = 0;
    }
}




void set_vlength( functor f )
{
  int i, j;
  int argument[ARGMAX];

  if ( !f ) return;
  set_vlength( f->nextfunct );
  /*
   *  if ( f->piped ) return;
   */
  f->fcc = tr_parameters.vlength;
  if (!f->adicity)
    tr_parameters.vlength++;
  else {
    for (i=0; i<f->adicity; i++) argument[i] = 0;
    do {
      j = vec_pos(f,argument);
      if (j >= tr_parameters.vlength) tr_parameters.vlength = j+1;
    }
    while (another_argument(f,argument));
  }
}



void set_forbidden( functor f )
{
  int i;
  int argument[ARGMAX];

  if ( !f ) return;
  set_forbidden( f->nextfunct );
  /*
   *  if ( f->piped ) return;
   */
  if (!f->adicity)
    tr_parameters.forbidden[f->fcc] = f->cut_it;
  else {
    for (i=0; i<f->adicity; i++) argument[i] = 0;
    do
      tr_parameters.forbidden[vec_pos(f,argument)] = f->cut_it;
    while (another_argument(f,argument));
  }
}



void set_sum(void)
{
  int i, j, k;
  functor f;
  char s[LEXLENGTH];

  for ( k = 0; sort[k].sname; k++ ) {
    if (sort[k].sum) free(sort[k].sum);
    sort[k].sum =
      (int*) malloc(((Integer->siz * sort[k].siz) + 1) * sizeof(int));
    sort[k].sum[0] = sort[k].siz;
    for ( i = 0; i < sort[k].siz; i++ )
      for ( j = 0; j < Integer->siz; j++ )
	sort[k].sum[(Integer->siz*i)+j+1] = i+j < sort[k].siz? i+j: sort[k].siz;
    strcpy(s,"+_"); strcat(s,sort[k].sname);
    f = newfunctor(s,Func,2);
    if (f->val_type != sort+k) {
      for (i=0; lexicon[i].word[0]; i++) printf("%s\n",lexicon[i].word);
      inerror("Incorrectly defined addition function",s);
    }
    f->fats[0] = Integer->siz;
    f->fats[1] = 1;
    f->table = sort[k].sum;
  }
}



void set_minus(void)
{
  int i, j, k;
  functor f;
  char s[LEXLENGTH];

  for ( k = 0; sort[k].sname; k++ ) {
    if (sort[k].minus) free(sort[k].minus);
    sort[k].minus =
      (int*) malloc(((Integer->siz * sort[k].siz) + 1) * sizeof(int));
    sort[k].minus[0] = sort[k].siz;
    for ( i = 0; i < sort[k].siz; i++ )
      for ( j = 0; j < Integer->siz; j++ )
	sort[k].minus[(Integer->siz*i)+j+1] = i >= j? i-j: sort[k].siz;
    sprintf(s,"-_%s",sort[k].sname);
    f = newfunctor(s,Func,2);
    if (f->val_type != sort+k) {
      for (i=0; lexicon[i].word[0]; i++) printf("%s\n",lexicon[i].word);
      inerror("Incorrectly defined subtraction function",s);
    }
    f->fats[0] = Integer->siz;
    f->fats[1] = 1;
    f->table = sort[k].minus;
  }
}


void set_pred(void)
{
  int i, k;
  functor f;
  char s[LEXLENGTH];

  for (k=0; sort[k].sname; k++) {
    if (sort[k].pred) free(sort[k].pred);
    sort[k].pred = (int*) malloc((sort[k].siz + 1) * sizeof(int));
    for (i=0; i <= sort[k].siz; i++)
      sort[k].pred[i] = (i>1? i-2: sort[k].siz);
    sprintf(s,"PRED_%s",sort[k].sname);
    f = newfunctor(s,Func,1);
    if (f->val_type != sort+k) {
      for (i=0; lexicon[i].word[0]; i++) printf("%s\n",lexicon[i].word);
      inerror("Incorrectly defined predecessor function",s);
    }
    f->table = sort[k].pred;
  }
}



void set_succ(void)
{
  int i, k;
  functor f;
  char s[LEXLENGTH];

  for (k=0; sort[k].sname; k++) {
    if (sort[k].succ) free(sort[k].succ);
    sort[k].succ = (int*) malloc((sort[k].siz + 1) * sizeof(int));
    for (i=0; i <= sort[k].siz; i++)
      sort[k].succ[i] = i? i: sort[k].siz;
    sprintf(s,"SUCC_%s",sort[k].sname);
    f = newfunctor(s,Func,1);
    if (f->val_type != sort+k) {
      for (i=0; lexicon[i].word[0]; i++) printf("%s\n",lexicon[i].word);
      inerror("Incorrectly defined successor function",s);
    }
    f->table = sort[k].succ;
  }
}



void set_E(void)
{
  int i, k;
  functor f;
  char s[LEXLENGTH];

  for ( k = 0; sort[k].sname; k++ ) {
    if (sort[k].Ex) free(sort[k].Ex);
    sort[k].Ex = (int*) malloc((sort[k].siz + 1) * sizeof(int));
    for ( i = 0; i <= sort[k].siz; i++ )
      sort[k].Ex[i] = i? true: false;
    sprintf(s,"EST_%s",sort[k].sname);
    f = newfunctor(s,Func,1);
    if (f->val_type != Bool) {
      for (i=0; lexicon[i].word[0]; i++) printf("%s\n",lexicon[i].word);
      inerror("Incorrectly defined existence function",s);
    }
    f->table = sort[k].Ex;
  }
}



void set_max(void)
{
  int k;

  for ( k = 0; sort[k].sname; k++ )
    sort[k].maxvalue = sort[k].siz - 1;
}



/*
 * This is a separate function because it is not just a subroutine of
 * set_cc (below). It is also called from functions in FINDER.c.
 */

void set_fats(functor f)
{
  int i;

  if (f->adicity)
    f->fats[f->adicity-1] = 1;
  for (i=f->adicity-2; i>= 0; i--)
    f->fats[i] = f->fats[i+1] * (f->arg_type[i+1]->siz);
}



void set_cc(void)
{
  functor f;

  for ( f = FN; f; f = f->nextfunct )
    set_fats(f);
}


void set_len(void)
{
  int i, k;

  for (k=0; sort[k].sname; k++)
    if (!sort[k].canonical) {
      if (sort[k].len) free(sort[k].len);
      sort[k].len = (int*) malloc(sort[k].siz * sizeof(int));
      for (i=0; i<sort[k].siz; i++)
	sort[k].len[i] = (i < 10? 1: 2);
    }
}


 /*
  *  This is extremely slow, of course, but as it's only used for
  *  debugging, speed is less important than obvious correctness.
  */

functor un_vec_pos(int x, int theArguments[])
{
  functor f;
  int i;

  for (f=FN; f; f = f->nextfunct)
    /*
     *    if (!f->piped) {
     */
    if (1) {
      for (i=0; i<f->adicity; i++) theArguments[i] = 0;
      do {
	if (vec_pos(f,theArguments) == x) return f;
      }
      while (another_argument(f,theArguments));
    }
  return 0;
}


void print_un_vec_pos(int x)
{
  functor f;
  int farg[ARGMAX];
  int i;

  f = un_vec_pos(x,farg);
  if (!f) printf("No function\n");
  else {
    printf("%s",f->fname);
    for (i=0; i<f->adicity; i++) {
      if (i) printf(","); else printf("(");
      printvalue(stdout,f->arg_type[i],farg[i],0);
    }
    if (i) printf(")");
  }
  fflush(stdout);
}



void print_un_vec_value(int x, int val)
{
  functor f;
  int farg[ARGMAX];

  f = un_vec_pos(x,farg);
  printvalue(stdout,f->val_type,val,0);
}



void report_constraint(ref thisref)
{
  intlist lp;
  int args[ARGMAX];
  functor f;
  int i;

  for (lp=thisref->cellvals; lp; lp = lp->p) {
    f = un_vec_pos(lp->v,args);
    if (!lp->positive) printf("not ");
    printf("%s(",f->fname);
    for (i=0; i<f->adicity; i++) {
      if (i) printf(",%d",args[i]);
      else printf("%d",args[i]);
    }
    printf(") = %d\n", lp->w);
  }
  if (thisref->softweight) printf("soft "); else printf("hard ");
  if (thisref->notop) printf("reduced");
  printf("\n");
  fflush(stdout);
}



void print_set( functor f )
{
  int i, j;
  int argument[ARGMAX];

  if (f->adicity > 2) {
    printf("\nAdicity of %s (fcc %d) greater than 2\n",f->fname,f->fcc);
    return;
  }
  printf("\nIndex for %s",f->fname);
  if ( !f->adicity ) {
    printf(" %d\n", vec_pos(f,argument));
    return;
  }
  puts("");
  if ( f->adicity == 1 ) {
    for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
      argument[0] = i;
      printf("%4d",vec_pos(f,argument));
    }
    argument[0] = i;
    printf("%5d\n",vec_pos(f,argument));
    return;
  }
  for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
    argument[0] = i;
    for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
      argument[1] = j;
      printf("%4d",vec_pos(f,argument));
    }
    argument[1] = j;
    printf("%5d",vec_pos(f,argument));
    puts("");
  }
  puts("");
  argument[0] = i;
  for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
    argument[1] = j;
    printf("%4d",vec_pos(f,argument));
  }
  argument[1] = j;
  printf("%5d",vec_pos(f,argument));
  puts("");
}



void set_max_penalty(int x)
{
  setting.target_bound = x;
}


/*
 * Ensure that the hidden funcions are at the end. Maintain the
 * internal order of the non-hidden functions, and that of the hidden
 * ones, by bubblesorting.
 */

void demote_hiddens(void)
{
  functor f;
  functor g;
  functor trail = 0;

  for (f=FN; f; f=f->nextfunct) {
    g = f->nextfunct;
    if (!(f->pmode) && f->cut_it && g && (g->pmode || !(g->cut_it))) {
      f->nextfunct = g->nextfunct;
      g->nextfunct = f;
      if (trail) trail->nextfunct = g;
      else FN = g;
      f = FN;
    }
    trail = f;
  }
  for (f=FN; f; f=f->nextfunct)
    if ((f->pmode || !f->cut_it) &&
	f->nextfunct &&	f->nextfunct->pmode && f->nextfunct->cut_it)
      f->nop = false;
}


boolean too_long_to_be_soft(int weight, int vars)
{
  if (vars+weight > setting.pretest_limit+2)
    return true;
  return false;
}
