/*
 * Before it can be used, a set must be initialised by means of a call
 * to either fullset, emptyset, singleton or assignset. If s does not
 * already exist, these allocate memory, which should be freed by
 * means of a call to freeset.
 */


#include "Fvntr.h"

typedef struct list_of_sets {
  set s;
  boolean allocated;
  struct list_of_sets *last;
  struct list_of_sets *next;
} SETLIST, *setlist;

static logical fullword = 0;
static int SSZ=0;              /* Number of words required for a set */
static setlist set_reservoir_start=0;
static setlist set_reservoir_end=0;
static int nsets = 0;


boolean init_sets_needed(int x)
{
  if (!set_reservoir_start) return true;
  if (!fullword) return true;
  if (x > SSZ) return true;
  return false;
}


void init_sets(int x)
{
  int i;

  for (i=0; i<WSZ; i++)
    fullword |= ((logical)1 << i);
  really_free_sets();
  SSZ = x;
  set_reservoir_start = (setlist) malloc(sizeof(SETLIST));
  set_reservoir_end = set_reservoir_start;
  set_reservoir_end->last = set_reservoir_end->next = 0;
  set_reservoir_end->s = malloc(SSZ*sizeof(logical));
  nsets++;
  set_reservoir_end->allocated = false;
}



set new_set()
{
  set s;
  setlist temp_setlist;

  if (set_reservoir_end->allocated) {
    temp_setlist = (setlist) malloc(sizeof(SETLIST));
    temp_setlist->s = (set) malloc(SSZ*sizeof(logical));
    nsets++;
    temp_setlist->allocated = false;
    temp_setlist->last = set_reservoir_end;
    set_reservoir_end->next = temp_setlist;
    set_reservoir_end = temp_setlist;
  }
  s = set_reservoir_end->s;
  set_reservoir_end->allocated = true;
  if (set_reservoir_start != set_reservoir_end) {
    temp_setlist = set_reservoir_end->last;
    set_reservoir_end->next = set_reservoir_start;
    set_reservoir_start->last = set_reservoir_end;
    set_reservoir_start = set_reservoir_end;
    set_reservoir_end = temp_setlist;
    set_reservoir_start->last = 0;
    set_reservoir_end->next = 0;
  }
  return s;
}



set fullset(set s)
{
  int i;

  if (!s) s = new_set();
  for (i=0; i<SSZ; i++)
    s[i] = fullword;
  return s;
}



set emptyset(set s)
{
  int i;
  set news;

  news = s? s: new_set();
  for (i=0; i<SSZ; i++)
    news[i] = 0;
  return news;
}



boolean isemptyset(set s)
{
  int i;

  for (i=0; i<SSZ; i++)
    if (s[i]) return false;
  return true;
}


set singleton(int x, set s)
{
  s = emptyset(s);
  intoset(x,s);
  return s;
}


boolean issingle(set s)
{
  int i,j;

  for (i=0; i<SSZ; i++)
    if (s[i]) {
      for (j=i+1; j<SSZ; j++)
	if (s[j]) return false;
      if (s[i] & (s[i]-1)) return false;
      return true;
    }
  return false;
}


set assignset(set dest, set source)
{
  int i;

  if (!dest) dest = new_set();
  for (i=0; i<SSZ; i++)
    dest[i] = source[i];
  return dest;
}



void intoset(int x, set s)
{
  s[x/WSZ] |= ((logical)1 << (x%WSZ));
}



void outofset(int x, set s)
{
  s[x/WSZ] &= ~((logical)1 << (x%WSZ));
}



boolean isinset(int x, set s)
{
  if (s[x/WSZ] & ((logical)1 << (x%WSZ)))
    return true;
  return false;
}



void intersect(set dest, set mask)
{
  int i;

  for (i=0; i<SSZ; i++)
    dest[i] &= mask[i];
}


void freeset(set s)
{
  setlist ptr;

  for (ptr=set_reservoir_start; ptr && ptr->allocated; ptr=ptr->next)
    if (ptr->s == s) {
      ptr->allocated = false;
      if (ptr == set_reservoir_end) return;
      ptr->next->last = ptr->last;
      if (ptr->last) ptr->last->next = ptr->next;
      if (ptr == set_reservoir_start)
	set_reservoir_start = set_reservoir_start->next;
      ptr->last = set_reservoir_end;
      ptr->last->next = ptr;
      ptr->next = 0;
      set_reservoir_end = ptr;
      return;
    }
  skipout("Attempt to de-allocate unallocated set",SKIP);
}

void print_reservoir()
{
  setlist ptr;

  for (ptr=set_reservoir_start; ptr; ptr=ptr->next) {
    printf("%d",(int)ptr->s);
    if (ptr->allocated) printf(" allocated\n");
    else printf(" unallocated\n");
  }
}


void really_free_sets(void)
{
  setlist ptr;
  void really_free_one_set(setlist ptr);

  for (ptr=set_reservoir_start; ptr; ptr=ptr->next) {
    really_free_one_set(ptr);
  }
  free(set_reservoir_end);
  set_reservoir_start = set_reservoir_end = 0;
}

void really_free_one_set(setlist ptr)
{
  if (ptr->s) {
    free(ptr->s);
    nsets--;
    ptr->s = 0;
  }
  if (ptr->last) {
    free(ptr->last);
    ptr->last =0;
  }
}
